/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include "tosdefs.h"
#include "mem.h"
#include "main.h"
#include "serial.h"
#include "screen.h"

#define SERIAL_DEV "/dev/ttyS1"
#define BOOTSECTOR 0

#define IGNORE_S(_addr) void STORE_B_ ## _addr (B v) {}
#define UNDEF_L(_addr) B LOAD_B_ ## _addr (void) { return 0xff; }
#define ZERO(_addr) B LOAD_B_ ## _addr (void) { return 0; }
#define NIB(_x) (((_x) & 0x0f)|0xf0)
#if BOOTSECTOR
/* custom bootsector with some 'disk' driver routines */
#include "bootsector.c"
#endif

/* This file contains the definitions of all the 'special' LOAD_* and STORE_*
 * functions for I/O-registers. The prototypes are contained in "iotab[12].c"
 * as generated from "iotab[12].tab" by "mktabl.pl".
 * 
 * For multi-byte accesses, the order will be lowest->highest byte, so e.g.
 * when writing to MWDATA (not implemented), STORE_B_ff8923 should trigger
 * the desired action.
 */

/* DMA registers */
static UW DMAfifo=0;	/* write to $8606.w */
static UW DMAstatus=0;	/* read from $8606.w */
static UW DMAxor=0;
static UB FDC_SC=0, HDC_SC=0;		/* sector count register */
/* (note that these could be implemented with a 4 byte array, but I chose not */
/* to do so for readability. FDC access shouldn't be very slow anyway ;-)     */
static UB FDC_CS=0, HDC_CS=0;		/* command/status register */
static UB FDC_T=0, HDC_T=0;		/* Track register */
static UB FDC_S=0, HDC_S=0;		/* Sector register */
static UB FDC_D=0, HDC_D=0;		/* Data register */
static UB DMAdiskctl=0;

extern volatile UL vbase;
int shiftmod;

/* generate a random Video Adress Counter value */
static UL gen_fake_vcount (void)
{
	return vbase + (rand() % 32000);
}

#define TG if (bank==1) return 0xff; time_gen();
/* save the time and date for later use */
static struct tm *tm;
static void time_gen (void)
{
	struct timeval tv;
	gettimeofday (&tv, NULL);
	tm = localtime (&(tv.tv_sec));
}

static void FDC (UB command)
{
	if ((command & -4) == 0)	/* RESTORE */
	{
		FDC_T = FDC_S = FDC_D = 0;
	}
}

static void HDC (UB command)
{
}

/* -------------------- ROUTINES FOR iotab1.tab -------------------- */

ZERO(ff8204)
ZERO(ff8206)
ZERO(ff8208)


B LOAD_B_ff8205(void)
{
	return (gen_fake_vcount() >> 16) & 0xff;
}

B LOAD_B_ff8207(void)
{
	return (gen_fake_vcount() >> 8) & 0xff;
}

B LOAD_B_ff8209(void)
{
	return gen_fake_vcount() & 0xff;
}

B LOAD_B_ff8604(void)
{
	return 0;
}

B LOAD_B_ff8605(void)
{
#if 0
	if (DMAfifo & 0x10)
	{
		return DMA_SC & 0xff; 
	}
	else
	{
		switch (DMAfifo & 6)
		{
		/* read status */
		case 0: return DMAstatus & 0xff; 
		/* read Track register */
		case 2: return FDC_T & 0xff; 
		/* read Sector register */
		case 4: return FDC_S & 0xff; 
		/* read Data register */
		case 6: return FDC_D & 0xff; 
		}
	}
#else
	return 0x04;
#endif
}

B LOAD_B_ff8606(void)
{
	return 0;
}

B LOAD_B_ff8607(void)
{
	return /* TODO */ 0x04;
}

#ifdef NO_AUDIO
B LOAD_B_ff8800(void)
{
	return /* TODO */ 0;
}

B LOAD_B_ff8802(void)
{
	return /* TODO */ 0;
}

UNDEF_L(ff8801)
UNDEF_L(ff8803)
#endif

UNDEF_L(ff8200)
IGNORE_S(ff8200)
IGNORE_S(ff8261)
IGNORE_S(ff8262)
IGNORE_S(ff8263)

void STORE_B_ff8260(B v)
{
	shiftmod = v;
	SM_B(MEM(0xff8260),v);
#if 0
	fprintf(stderr,"Shiftmode set\n");
#endif
	machine.screen_shifter();
}

void set_vbase (UL n)
{
	SM_B(MEM(0xff8201),(n>>16)&0xff);
	SM_B(MEM(0xff8203),(n>>8)&0xff);
#if STE
	SM_B(MEM(0xff820d),n&0xff);
#endif
	vbase = n;
}

void STORE_B_ff8201(B v)
{
	SM_B(MEM(0xff8201),v);
	vbase &= 0xff00ffff;
	vbase |= (v & 0xff)<<16;
#if 0
	fprintf(stderr,"vbase modified from %lx (->%lx)\n",pc,vbase);
#endif
}

UNDEF_L(ff8202)
IGNORE_S(ff8202)

void STORE_B_ff8203(B v)
{
	SM_B(MEM(0xff8203),v);
	vbase &= 0xffff00ff;
	vbase |= (v & 0xff)<<8;
#if 0
	fprintf(stderr,"vbase modified from %lx (->%lx)\n",pc,vbase);
#endif
}

void STORE_B_ff820d(B v)
{
#if STE
	SM_B(MEM(0xff820d),v);
	vbase &= 0xffffff00;
	vbase |= (v & 0xff);
#endif
}

B LOAD_B_ff820d(void)
{
	return 0xff;
}

void STORE_B_ff820a(B v)
{
#if 0
	char *n;
	/* just for fun... */
	switch (v & 3)
	{
	case 0:	n = "STemu, 60Hz, internal sync."; break;
	case 1:	n = "STemu, 60Hz, external sync."; break;
	case 2:	n = "STemu, 50Hz, internal sync."; break;
	case 3:	n = "STemu, 50Hz, external sync."; break;
	}
	set_window_name (n);
#endif
}

void STORE_B_ff8604(B v)
{
	/* nothing */
}

void STORE_B_ff8605(B v)
{
	DMAdiskctl = v & 0xff;
}

void STORE_B_ff8606(B v)
{
	DMAfifo &= 0xff;
	DMAfifo |= v<<8;
}

void STORE_B_ff8607(B v)
{
#if 0
	DMAfifo &= 0xff00;
	DMAfifo |= v & 0xff;
	/* if the Sector-Control Register was selected, it's a 'Kippeln' (?) */
	if (DMAfifo & 0x10)
	{
		DMAxor ^= DMAfifo;
	}
	else
	{
		switch (DMAfifo & 14)
		{
		case 8: HDC(HDC_CS = DMAdiskctl); break;
		case 10: HDC_T = DMAdiskctl; break;
		case 12: HDC_S = DMAdiskctl; break;
		case 14: HDC_D = DMAdiskctl; break;
		case 0:	FDC(FDC_CS = DMAdiskctl); break;	
		case 2: FDC_T = DMAdiskctl; break;
		case 4: FDC_S = DMAdiskctl; break;
		case 6: FDC_D = DMAdiskctl; break;
		}
	}
#else
#if BOOTSECTOR
	if ((v&0xff) == 0x88 && DMAdiskctl == 1)
	{
		UL x = (LM_UB(MEM(0xff8609))<<16)|(LM_UB(MEM(0xff860b))<<8)
				|LM_UB(MEM(0xff860d));
		bcopy(bootsec, MEM(x), 512);
	}
#endif
#endif
}

#ifdef NO_AUDIO
IGNORE_S(ff8800)
IGNORE_S(ff8801)
IGNORE_S(ff8802)
IGNORE_S(ff8803)
#endif

B LOAD_B_ff8920(void) {return 0;}
B LOAD_B_ff8910(void) {return 0;}
B LOAD_B_ff8900(void) {return 0;}
B LOAD_B_ff8921(void) {return 0;}
B LOAD_B_ff8911(void) {return 0;}
B LOAD_B_ff8901(void) {return 0;}
B LOAD_B_ff8922(void) {return 0;}
B LOAD_B_ff8912(void) {return 0;}
B LOAD_B_ff8902(void) {return 0;}
B LOAD_B_ff8923(void) {return 0;}
B LOAD_B_ff8913(void) {return 0;}
B LOAD_B_ff8903(void) {return 0;}
B LOAD_B_ff8924(void) {return 0;}
B LOAD_B_ff8904(void) {return 0;}
B LOAD_B_ff8925(void) {return 0;}
B LOAD_B_ff8905(void) {return 0;}
B LOAD_B_ff8906(void) {return 0;}
B LOAD_B_ff8907(void) {return 0;}
B LOAD_B_ff8908(void) {return 0;}
B LOAD_B_ff8909(void) {return 0;}
B LOAD_B_ff890a(void) {return 0;}
B LOAD_B_ff890b(void) {return 0;}
B LOAD_B_ff890c(void) {return 0;}
B LOAD_B_ff890d(void) {return 0;}
B LOAD_B_ff890e(void) {return 0;}
B LOAD_B_ff890f(void) {return 0;}
/* -------------------- ROUTINES FOR iotab2.tab -------------------- */
B LOAD_B_fffc00(void)
{
	return LM_B(MEM(0xfffc00)) | 2;
}

B LOAD_B_fffc02(void)
{
	return ikbd_readbuf();
}

B LOAD_B_fffc04(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffc06(void)
{
	return /* TODO */ 1;
}
void STORE_B_fffc00(B v)
{
	/* TODO */
}

void STORE_B_fffc02(B v)
{
	UB x;
	/* send to IKBD (TODO) */
	x = LM_UB(MEM(0xfffc02));
	ikbd_send_byte(v);
	x |= 2;
	SM_UB(MEM(0xfffc02), x);
}

void STORE_B_fffc04(B v)
{
	/* TODO */
}

void STORE_B_fffc06(B v)
{
	/* TODO */
}

static int bank=0;
IGNORE_S(fffc20)
IGNORE_S(fffc22)
IGNORE_S(fffc24)
IGNORE_S(fffc26)
IGNORE_S(fffc28)
IGNORE_S(fffc2a)
IGNORE_S(fffc2c)
IGNORE_S(fffc2e)
IGNORE_S(fffc30)
IGNORE_S(fffc32)
IGNORE_S(fffc34)
IGNORE_S(fffc36)
IGNORE_S(fffc38)
IGNORE_S(fffc3a)
IGNORE_S(fffc3c)
IGNORE_S(fffc3e)
UNDEF_L(fffc20)
UNDEF_L(fffc22)
UNDEF_L(fffc24)
UNDEF_L(fffc26)
UNDEF_L(fffc28)
UNDEF_L(fffc2a)
UNDEF_L(fffc2c)
UNDEF_L(fffc2e)
UNDEF_L(fffc30)
UNDEF_L(fffc32)
UNDEF_L(fffc34)
UNDEF_L(fffc36)
UNDEF_L(fffc38)
UNDEF_L(fffc3a)
UNDEF_L(fffc3c)
UNDEF_L(fffc3e)

/* Normally, we would set the clock according to what's written to these
   addresses. But since we use the Unix system clock, we don't want (and,
   often, we can't) change it. */

IGNORE_S(fffc21)
IGNORE_S(fffc23)
/* c25, c27 below */
IGNORE_S(fffc29)
IGNORE_S(fffc2b)
IGNORE_S(fffc2d)
IGNORE_S(fffc2f)
IGNORE_S(fffc31)
IGNORE_S(fffc33)
IGNORE_S(fffc37)
IGNORE_S(fffc39)
/* c3b below */
IGNORE_S(fffc3d)
IGNORE_S(fffc3f)

static int mode24=1;
void STORE_B_fffc35 (B v)
{
	if (bank == 1) mode24 = v & 1;
}

static B fake_am, fake_amz;
void STORE_B_fffc25 (B v)
{
	/* TOS uses this... */
	if (bank == 1) fake_am = NIB(v);
	/* else ignore */
}
void STORE_B_fffc27 (B v)
{
	if (bank == 1) fake_amz = NIB(v);
}


static B c3b;
void STORE_B_fffc3b (B v)
{
	bank = v & 1;	/* the other 2 bits are ignored */	
	c3b = NIB(v);
}

#define TFUNC(_addr,_ret) B LOAD_B_ ## _addr (void) {TG return NIB(_ret);}
TFUNC(fffc21,tm->tm_sec % 10)
TFUNC(fffc23,tm->tm_sec / 10)
TFUNC(fffc29,tm->tm_hour % 10)
TFUNC(fffc2d,tm->tm_wday)
TFUNC(fffc2f,tm->tm_mday % 10)
TFUNC(fffc31,tm->tm_mday / 10)
TFUNC(fffc33,(tm->tm_mon+1) % 10)
TFUNC(fffc35,(tm->tm_mon+1) / 10)
TFUNC(fffc37,tm->tm_year % 10)
TFUNC(fffc39,(tm->tm_year-80) / 10)	

B LOAD_B_fffc3b (void)
{
	return c3b;
}

B LOAD_B_fffc2b(void)
{
	TG 
	if (mode24) return tm->tm_hour / 10;
	else
	{
		return (1<<(tm->tm_hour == 0 || tm->tm_hour >= 13))
				|((tm->tm_hour % 12)==0) || ((tm->tm_hour % 12)>=10);
	}
}

B LOAD_B_fffc25(void)
{
	if (bank == 1) return fake_am;
	time_gen();
	return NIB(tm->tm_min % 10);
}
B LOAD_B_fffc27(void)
{
	if (bank == 1) return fake_amz;
	time_gen();
	return NIB(tm->tm_min / 10);
}

B LOAD_B_fffc3d(void)
{
	return NIB(0);
}

UNDEF_L(fffc3f)
