/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "missing.h"
#include "error.h"
#include "gemdos.h"
#include "main.h"
#include "mem.h"
#include "screen.h"
#include "version.h"

#define STONXRC ".stonxrc"
#define EQ(_x,_y) (is_rc ? (strcmp(_x,_y)==0) : (strcmp(_x,"-"_y)==0) )

int gr_mode=MODE_MONO;
int swidth=640, sheight=400;
extern int scr_width, scr_height, scr_planes;
int custom_screensize=0;
int boot_dev=0;
int vblpt=4;	/* interrupts per VBL */
int refresh=2;	/* VBLs per refresh */
int hz200pt=1;	/* interrupts per timer C interrupt */
long tv_usec=10000;
int verbose=1;
int vdi=0;
int vdi_mode=0;
int audio=0;
int scanlines=200;
int warmboot=0;
int priv_cmap=0;
int redirect_cconws=0;
int fix_screen=0;
extern char *serial_dev;
extern char *parallel_dev;
extern int shmflag;
int got_drive=0;
int chunky=1;
extern char *kdefsfile;
Machine_Specific machine;

void process_args (int argc, char *argv[], int is_rc)
{
	int i, u=0;
	for (i=1; i<argc; i++)
	{
		char *x = argv[i];
		if (EQ(x,"mono")) gr_mode = MODE_MONO;
		else if (EQ(x,"color")) gr_mode = MODE_COLOR;
		else if (EQ(x,"colour")) gr_mode = MODE_COLOR;
		else if (EQ(x,"size"))
		{
			if (i == argc-1)
				error("size needs an argument of the form <width>x<height>\n");
			sscanf(argv[++i],"%dx%d",&swidth,&sheight);
			if (swidth <= 0 || sheight <= 0)
				error("invalid argument to -size\n");
			swidth&=0xFFFFFFE0;
			sheight&=0xFFFFFFE0;
			custom_screensize=1;
		}
		else if (EQ(x,"vbl"))
		{
			int v;
			if (i == argc-1 || (v = atoi(argv[++i])) <= 0)
				error("vbl needs a numeric argument > 0\n");
			vblpt=v;
		}
		else if (EQ(x, "q"))
		{
			verbose=0;
		}
#ifndef __BEOS__
		else if (EQ(x,"private"))
		{
			priv_cmap=1;
		}
#endif
#ifdef	USE_SVGALIB
		else if (EQ(x,"svga"))
		{
			machine.screen_open = svga_screen_open;
			machine.screen_close = svga_screen_close;
			machine.screen_shifter = svga_screen_shifter;
			machine.init_keys = svga_init_keys;
			
		}
#endif
		else if (EQ(x,"nochunky"))
		{
			chunky=0;
		}
#ifndef __BEOS__
		else if (EQ(x,"kmap"))
		{
			if (i == argc-1)
				error("kmap needs a file name as argument (default: Keysyms)\n");
			kdefsfile = strdup(argv[++i]);
		}
#endif
		else if (EQ(x,"cconws"))
		{
			redirect_cconws=1;
		}
		else if (EQ(x, "refresh"))
		{
			int t;
			if (i == argc-1 || (t = atoi(argv[++i])) <= 0)
				error("timer-c needs a numeric argument > 0\n");
			refresh=t;
		}
		else if (EQ(x,"timer-c"))
		{
			int t;
			if (i == argc-1 || (t = atoi(argv[++i])) <= 0)
				error("timer-c needs a numeric argument > 0\n");
			hz200pt=t;
		}
		else if (EQ(x,"usec"))
		{
			long x;
			if (i == argc-1 || (x = atol(argv[++i])) <= 0)
				error("usec needs a numeric argument > 0\n");
			tv_usec = x;
		}
#ifndef __BEOS__
		else if (EQ(x,"vdi"))
		{
			vdi_mode = vdi = 1;
		}
		else if (EQ(x,"scanlines"))
		{
			int x;
			if (i == argc-1 || (x = atoi(argv[++i])) <= 0)
				error("scanlines needs a numeric argument > 0\n");
			scanlines = x;
		}
#endif
		else if (EQ(x,"disk"))
		{
			char d;
			char fname[256];	/* use value from limits.h? */
			if (i == argc-1 || sscanf(argv[++i],"%c:%s", &d, fname) < 2)
				error("disk needs an argument of the form <X>:<File> where"
					" <X> is a drive letter\n");
			if (!add_drive (d, fname)) u=1;
			else got_drive=1;
		}
		else if (EQ(x,"fs"))
		{
			char d;
			char fname[256];	
			if (i == argc-1 || sscanf(argv[++i],"%c:%s", &d, fname) < 2)
				error("fs needs an argument of the form <X>:<directory> where"
					" <X> is a drive letter\n");
			if (!add_gemdos_drive (d, fname)) u=1;
			else got_drive=1;
		}
		else if (EQ(x,"serial"))
		{
			if (i == argc-1)
				error("serial needs a device as argument (/dev/ttyS1 for example)\n");
			serial_dev = strdup(argv[++i]);
		}
		else if (EQ(x,"para"))
		{
			if (i == argc-1)
				error("para needs a device as argument (/dev/lp0 for example)\n");
			parallel_dev = strdup(argv[++i]);
		}
#ifndef __BEOS__
		else if (EQ(x,"noaudio"))
		{
			audio=0;
		}
		else if (EQ(x,"noshm"))
		{
			shmflag=0;
		}
#endif
		else if (EQ(x,"warmboot"))
		{
			warmboot=1;
		}
		else if (EQ(x,"h") || EQ(x,"?") || EQ(x,"-help"))
			u=1;
		else
		{
			fprintf(stderr, "unknown option %s\n", argv[i]);
			u=1;
		}

	}
	if (u)
	{
		if (is_rc)
			fprintf(stderr,"Error in "STONXRC", please fix!\n");
		else
		(void)fprintf (stderr, "usage: %s [<options>]\n"
			"<options> = \n"
" -q                        Be Quiet\n"
" -size <W>x<H>             Define custom screensize.\n"
" -disk <X>:<File>          Load diskfile <File> as Drive <X>:\n"
" -fs <X>:<Directory>       Map <Directory> to Drive <X>:\n"
" -colo[u]r|-mono           Set `monitor type'\n"
" -usec <X>                 Set <X> microsecond interval timer\n"
" -vbl <X>                  Set <X> * usec VBL interval\n"
#if DOUBLE_HZ200
" -timer-c <X>              Set <X> * usec / 2 Timer C interval\n"
#else
" -timer-c <X>              Set <X> * usec Timer C interval\n"
#endif
" -refresh <X>              Set <X> * VBL refresh interval\n"
" -para <file>              Use <file> as the parallel port device\n"
" -serial <file>            Use <file> as the serial port device\n"
#ifndef __BEOS__
" -kmap   <file>            Load the Keyboard mappings from <file>\n"
" -noaudio                  Disable Audio driver\n"
" -noshm                    Disable XShm extension\n"
#endif
" -nochunky                 Disable `chunky' update mode\n"
" -warmboot                 Attempt to simulate a warm-boot\n"
#ifndef __BEOS__
" -private                  Use private colormap (speeds -color up)\n"
#endif
" -cconws                   Redirect GEMDOS Cconws() function to terminal\n"
#ifdef	USE_SVGALIB
" -svga                     Use Linux SVGALIB instead of X\n"	
#endif
#ifndef __BEOS__
" -vdi                      Use xlib-VDI\n\n"
#endif
"Example: %s -disk a:Disk -color -usec 10000 -vbl 10 -refresh 2 -timer-c 1\n"
"will use `Disk' for drive A:, boot in color mode, and run Timer C at "
#if DOUBLE_HZ200
"200Hz"
#else
"100Hz"
#endif
",\nVBLs at "
"10Hz"
", and refresh with "
"5Hz"
", if your system allows interval timers with\n10 ms timer intervals (see setitimer(2))\n\n", argv[0], argv[0]);
		exit(0);
	}
}

void process_stonxrc(void)
{	char *home;
    	if ((home=getenv("HOME")) != NULL)
	{	FILE *rc;	
		char stonxrc[256];
		strcpy (stonxrc,home);
		strcat (stonxrc,"/"STONXRC);
		if ((rc=fopen(stonxrc,"r")) != NULL)
		{	int i,c=1;	
			char *args[64];
			for (i=0;i<64;i++) args[i] = malloc(sizeof(char)*256);
			while (fscanf(rc,"%s",args[c]) != EOF) {
				if (args[c][0] == '#') {
					int ch;
					while ((ch = fgetc(rc)) != EOF &&
						ch != '\n')
						;
					continue;
				}
				c++;
			}
			fclose(rc);
			fprintf (stderr, "Using configuration file:%s\n", stonxrc);
			process_args (c, args,1);
			for (i=1 ;i<64;i++) free(args[i]);
		}
	}
}

void print_timing(void)
{
#ifdef DOUBLE_HZ200
#define MVAL0 2000000.0
#else
#define MVAL0 1000000.0
#endif
	if (verbose) 
	{	fprintf (stderr, "%ld microseconds interval timer\n", tv_usec);
		fprintf (stderr, "%.0f Hz Timer C, ", (double) MVAL0/(tv_usec*hz200pt));
		fprintf (stderr, "%.0f Hz VBLs and %.0f Hz screen update\n",
				(double)1000000.0/(vblpt*tv_usec),
				(double)1000000.0/(vblpt*refresh*tv_usec));
	}
}

extern void InitApp(int argc, char *argv[]);



int main (int argc, char *argv[])
{
	InitApp(argc, argv);
	
	DeleteApp();
	return 0;
}

void kill_beos()
{
	kill_hardware();
#if MONITOR
	kill_monitor();
#endif
	machine.screen_close();
	if (parallel_dev != NULL) done_parallel();
	if (audio) audio_close();
	exit_mem();
	exit_debug();
}

int
main_beos (int argc, char *argv[])
{
	flags = 0;

	machine.screen_open = x_screen_open;
	machine.screen_close = x_screen_close;
	machine.screen_shifter = x_screen_shifter;
	machine.init_keys = x_init_keys;

	printf ("STonBeOS Version " VERSION "\n(c) 1995 Marinos Yannikos, Martin D. Griffiths and Tinic Urou\n\n"
	"*** This program is free software, and comes with NO WARRANTY! ***\n\n"
	"**************** Read the file COPYING for details ***************\n\n");
	process_stonxrc ();
	process_args (argc, argv, 0);
	if (gr_mode == MODE_COLOR && vdi)
	{
		fprintf(stderr,"Warning: -colour -vdi implies -private!\n");
		priv_cmap=1;
	}
	if (!custom_screensize && gr_mode == MODE_COLOR)
	{
		swidth=320;
		sheight=200;
		fix_screen=0;
	}
	if ((gr_mode == MODE_MONO && (swidth != 640 || sheight != 400)) 
		|| gr_mode == MODE_COLOR && (swidth != 320 || sheight != 200))
		fix_screen=1;
	if (vdi && custom_screensize)
	{
		int i;
		for (i=0; i<3; i++)
		{
			scr_def[i].w = swidth;
			scr_def[i].h = sheight;
		}
	}
	else if (gr_mode == MODE_COLOR)
	{
		scr_def[0].w = swidth;
		scr_def[1].w = scr_def[2].w = swidth*2;
		scr_def[0].h = scr_def[1].h = sheight;
		scr_def[2].h = sheight*2;
	}
	else /* mode == MONO or no custom screen size or no VDI */
	{
		scr_def[0].w = swidth/2;
		scr_def[1].w = scr_def[2].w = swidth;
		scr_def[0].h = scr_def[1].h = sheight/2;
		scr_def[2].h = sheight;
	}
	/* need these for init_mem() */
	if (gr_mode == MODE_COLOR)
	{
		scr_width=scr_def[0].w;
		scr_height=scr_def[0].h;
		scr_planes=4;
	}
	else
	{
		scr_width=scr_def[2].w;
		scr_height=scr_def[2].h;
		scr_planes=1;
	}
	print_timing();
	if (!got_drive)
	{
		char path[500];
		getcwd(path,500);
		(void)fprintf (stderr, "No drives specified - using default: `-fs C:%s'\n", path);
		add_gemdos_drive ('C',path);
	}
	init_debug();
	init_ui();
	init_mem();
	load_file ("tos.img",MEM(TOSSTART));
#if CARTRIDGE
	load_file ("cartridge.img", MEM(0xfa0000));
#endif
	machine.screen_open();
	if (audio) audio_open();
	init_hardware();
	init_ikbd();
#if MONITOR
	init_monitor();
#endif
	if (parallel_dev != NULL) init_parallel();
	if (serial_dev != NULL) init_serial();
#if 0
	load_file ("myrom.img", MEM(0x100000));
#endif
	init_cpu();
}

