/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "debug.h"
#include "main.h"
#include "tosdefs.h"
#include "gemdos.h"
#include "xlib_vdi.h"

#define SERIAL_DEV "/dev/ttyS1"

#define FUNC_OPCODE 0xa0ff
#define BPB_OFF 0xfa4000

#define	FLG_RDONLY 	1
#define FLG_SPECIAL	2

extern int verbose,fix_screen,scr_width,scr_height,scr_planes;
extern volatile UL vbase;
int drive_bits;
int drive_fd[26];
static int drive_flags[26];

extern void Vdi(UL as);
extern void Init_Linea(UL as);

/* Add a drive to the system; this must happen before TOS is booted */
int add_drive (char drive, char *fname)
{
	int fd, tried=0, r;
	struct stat st;
	char msg[1000];
	int d = toupper(drive & 0xff) - 'A';
	if (d < 0 || d >= 26)
	{
		(void)fprintf (stderr, "invalid drive specification `%c'\n",drive);
		return 0;
	}
	if (drive_bits & (1<<d))
	{
		(void)fprintf (stderr, "drive `%c' already specified\n", drive);
		return 0;
	}
	fd = open (fname, O_RDWR);
nexttry:
	if (fd < 0)
	{
		if (errno == EROFS)
		{
			fd = open (fname, O_RDONLY);
			if (fd > 0)
			{
				fprintf (stderr, "Warning: file %s is read-only\n", fname);
				drive_flags[d] |= FLG_RDONLY;
				goto opened;
			}
		}
		sprintf (msg, "can't open file %s (%d)", fname, errno);
		perror (msg);
		exit(2);
	}
opened:
	r = fstat(fd, &st);
	if (r < 0)
	{
		sprintf (msg, "Warning: can't stat file %s", fname);
		perror (msg);
	}
	else
	{
		if (!S_ISREG(st.st_mode))
		{
			fprintf (stderr, "Note: %s is a special file - it may or may not"
					" work as desired...\n", fname);
			drive_flags[d] |= FLG_SPECIAL;
			if (S_ISBLK(st.st_mode))
			{
				fprintf (stderr, "Note: if you want to emulate a `real' floppy disk drive, you should use the\ncorresponding raw device instead!\n");
			}
		}
		if (st.st_size == 0)
		{
			fprintf (stderr, "Warning: file %s has size 0!\n", fname);
		}
	}		
	drive_bits |= 1<<d;
	drive_fd[d] = fd;
	return 1;
}

static void write_native (UL as)
{
	UL addr = LM_UL(MEM(as));
	if (verbose)
		fputs (MEM(addr), stderr);
}

static void disk_rw (UL as)
{
	W rwflag;
	UL buf;
	unsigned int count, recno, dev;
	rwflag = LM_W(MEM(as));
	buf = LM_UL(MEM(as+2));
	count = LM_UW(MEM(as+6));
	recno = LM_UW(MEM(as+8));
	dev = LM_UW(MEM(as+10));
	DEBUG && D ("Rwabs(%d,$%x,%d,%d,%d), %lx, %d\n", rwflag, buf, count, recno, dev,drive_bits,boot_dev);
	if (dev < 0 || dev >= 26 || !(drive_bits & (1<<dev)) || drive_fd[dev]<0)
	{
		DREG(0) = -15;
		return;
	}
	if ((drive_flags[dev] & FLG_RDONLY) && (rwflag & 1))
	{
		DREG(0) = -13;
		return;
	}
	rwflag &= 1;	/* media change not yet implemented... */
	if (lseek (drive_fd[dev], recno*512, SEEK_SET) < 0)
	{
		perror ("lseek() failed");
		DREG(0) = -6;	/* E_SEEK is probably inappropriate */
		return;
	}
	if (rwflag == 0)
	{
		if (read (drive_fd[dev], MEM(buf), count*512) < 0)
		{
			perror ("read() failed (not 9/18 sectors/track?)");
			DREG(0) = -11;
			return;
		}
	}		
	else
	{
		/* we should not lseek() if the following is true... */
		if (drive_flags[dev] & FLG_RDONLY)
		{
			DREG(0) = -13;
			return;
		}
		if (write (drive_fd[dev], MEM(buf), count*512) < 0)
		{
			perror ("write() failed (not 9/18 sectors/track?)");
			DREG(0) = -10;
			return;
		}
	}
	DREG(0) = 0;
	return;
}

/* Is it illegal to return a BPB address which is read-only???
 */
static void disk_bpb (UL as)	/* drive # ignored! */
{
	UB bootsec[512], *d;
	UB *x = MEM(BPB_OFF);
	UW *wx = (UW *)x;
	int dev = LM_W(MEM(as));
	off_t old;
	DEBUG && D("Getting BPB for drive %d\n",dev);
	if (dev < 0 || dev >= 26 || !(drive_bits & (1<<dev)) || drive_fd[dev]<0)
	{
		DREG(0) = 0;
		return;
	}
	old = lseek(drive_fd[dev], 0, SEEK_CUR);
	/* TODO: check return values... :-( */
	lseek(drive_fd[dev], 0, SEEK_SET);
	read(drive_fd[dev], bootsec, 512);
	lseek(drive_fd[dev], old, SEEK_SET);
	d = bootsec;
	x[0] = d[0xc];
	x[1] = d[0xb];
	x[2] = 0;	/* BUG */
	x[3] = d[0xd];
	SM_W(&wx[2],LM_W(&wx[1]) * LM_W(&wx[0]));
	SM_W(&wx[3],(d[0x11] + d[0x12]*256)/16);
	x[8] = d[0x17];
	x[9] = d[0x16];
	SM_W(&wx[5],LM_W(&wx[4])+1);
	SM_W(&wx[6],LM_W(&wx[5])+LM_W(&wx[4])+LM_W(&wx[3]));
	SM_W(&wx[7],((d[0x13] + d[0x14]*256)-LM_W(&wx[6]))/LM_W(&wx[1]));
	SM_W(&wx[8],(LM_W(&wx[7])>0xfef));	/* WRONG!? */
	DREG(0) = BPB_OFF;
}

static void Init_Screen (UL as)
{
	int mem_wanted;
	UL memtop,phystop,new_top;
	/* this needs fixing (FIXME), will break reset-resident ramdisks etc. (?) */
#if SMALL
#define PHYSTOP 0x400000
#else
#define PHYSTOP 0xe00000
#endif
	memtop=LM_UL(MEM(0x436));
	fprintf(stderr,"Manual memory configuration for custom screen size...\n");
	fprintf(stderr,"memtop=%08lx\n",memtop);
	mem_wanted = (scr_height*scr_width*scr_planes)/8;
	new_top = (PHYSTOP-mem_wanted)&-1024;
#if 0
	SM_UL(MEM(0x436),new_top);
	SM_UL(MEM(0x44e),new_top);
	SM_UL(MEM(0x42e),new_top);
	SM_UL(MEM(0x420),0x752019f3);
	SM_UL(MEM(0x43a),0x237698aa);
	SM_UL(MEM(0x51a),0x5555aaaa); 
#if SMALL
	SM_UB(MEM(0x424),10);
#else
	SM_UB(MEM(0x424),15);
#endif
	SM_UL(MEM(0x4ba),16000);
	vbase = new_top;
#endif
	fprintf(stderr,"new: %08lx\n",new_top);
}

static void Initialize (UL as)
{
	extern int gemdos_ok;
	extern int vdi_ok,linea_ok;
	gemdos_ok=0;
	vdi_ok=0;
}

static void disk_init (UL as)
{
	extern int linea_ok;
	int b=boot_dev;
	if (b == 0 && !(drive_bits & 1))
		while (!((drive_bits >> b)&1)) b++;
	boot_dev = b;
#if 0
	if ((drive_bits & 1)) drive_bits |= 2;
#endif
	SM_W(ADDR(0x446), boot_dev);
	SM_W(ADDR(0x4a6), 0); /* _nflops may be different if we have FDC emul.*/
	SM_L(ADDR(0x4c2), drive_bits);
	fprintf (stderr, "boot_dev = %d cmdload = %d\n",boot_dev, LM_W(ADDR(0x482)));	
	if (!linea_ok)
	{
		linea_init();
		linea_ok=1;
	}
}

/* -------------- Serial Port functions ------------- */
/* A0 is the argument stack here... */
static void Bconin_AUX (UL as)
{
	int td;
	char c;
	fprintf (stderr, "Bconin_AUX\n");
	td = open (SERIAL_DEV, O_RDONLY);
	if (td < 0)
	{
		perror ("open " SERIAL_DEV);
		DREG(0) = -1;
		return;
	}
	if (read (td, &c, 1) < 0)
	{
		perror ("read " SERIAL_DEV);
		DREG(0) = -1;
		return;
	}
	close(td);
}

static void Bconout_AUX (UL as)
{
	int td;
	char c = LM_W(MEM(AREG(0)+4));
	fprintf (stderr, "Bconout_AUX (%c = %d)\n",c,(int)c&0xff);
	td = open (SERIAL_DEV, O_WRONLY);
	if (td < 0)
	{
		perror ("open " SERIAL_DEV);
		DREG(0) = -1;
		return;
	}
	if (write (td, &c, 1) < 0)
	{
		perror ("write " SERIAL_DEV);
		DREG(0) = -1;
		return;
	}
	close(td);
}

static void Bconstat_AUX (UL as)
{
	fprintf (stderr, "Bconstat_AUX\n");
}

static void Bcostat_AUX (UL as)
{
	fprintf (stderr, "Bcostat_AUX\n");
}

static void Rsconf (UL as)
{
	fprintf (stderr, "Rsconf\n");
}

static void move_native (UL as)
{
	fprintf (stderr, "%lx -> %lx\n", DREG(0), AREG(0));
	SM_L(MEM(AREG(0)),DREG(0));
}

void (*native_funcs[])(UL as) = 
{
	write_native,	/* print a string in the 'native' window 	*/
	disk_rw,		/* _hdv_rw native routine 					*/
	disk_bpb,		/* _hdv_bpb 								*/
	disk_init,	
	move_native,	/* write to ROM */
	Gemdos,
	vdi_post,
	linea_post,
	Vdi,
	Initialize,
	Init_Linea,
	Init_Screen,
};

void call_native (UL as, UL func)
{
	if (func >= sizeof(native_funcs)/sizeof(native_funcs[0])) return;
	native_funcs[func](as);
}
