/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */

#include "defs.h"
#include "tosdefs.h"
#include "debug.h"
#include "ikbd.h"
#include "main.h"
#include "missing.h"
#include "onexit.h"
#include "screen.h"
#include "xlib_vdi.h"
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/times.h>
#include <time.h>
#include <KernelKit.h>

SCRDEF scr_def[3];
int shmflag = 1; 
char *kdefsfile = "Keysyms";
char mapcol[COLS];

sem_id b1,b2;

static int old_shiftmod=-1;	/* last known shift mode, used to detect changes */
UW old_color[16];	/* last known color palette */

/* This function is called periodically (from a signal handler) to check what's
 * new with the shifter mode, colors, and update ("shift") the screen. It
 * should probably rewritten to check whether it's necessary to update the
 * screen, since for color modes this is a very costly operation...
 *
 * The Shifter status information is changed in hw.c, where the hardware regs
 * are read.
 */

extern void change_colors (void);

static void change_mode (void)
{
	scr_width = scr_def[shiftmod].w;
	scr_height = scr_def[shiftmod].h;
	scr_planes=4>>shiftmod;
	fprintf(stderr,"Screen is now %dx%d, shift mode %d\n",scr_width,scr_height,shiftmod);
	ResizeWindow();
}

extern void Lock();
extern void Unlock();
extern void DirectDraw(int x1, int y1, int x2, int y2);

static void show_screen (void)
{
	extern unsigned char *bits;
	extern unsigned long xmod;
	int w, h;

	while(acquire_sem(b1)==B_NO_ERROR)
	{	
		int xzoom,yzoom=1;
		w=scr_width;
		h=scr_height;
		switch (old_shiftmod)
		{
			case 0:
				st16c_to_z (MEM(vbase), bits);
				break;
			case 1:
				yzoom=2;
				st4c_to_z (MEM(vbase), bits);
				break;
			case 2:
				stmono_to_xy(bits, MEM(vbase));
				break;
		}
	
		if (chunky)
		{
			int i;
			Lock();
			for (i=0; i<h/CHUNK_LINES; i++)
				if (chunk_age[i]==1)
				{
					DirectDraw
					(
						0, yzoom*(i+0)*CHUNK_LINES,
						w, yzoom*(i+1)*CHUNK_LINES
					);
#if BENCH_REFRESH
					chunks_drawn++;
#endif
				}
#if BENCH_REFRESH
			chunks_seen+=h/CHUNK_LINES;
#endif
			Unlock();
		}
		else Update();
	
		release_sem(b2);
	}
}

void x_screen_close (void)
{
	CloseWindow();
	delete_sem(b1);
	delete_sem(b2);
}

void x_screen_open (void)
{
	x_init_keys();

	b1=create_sem(0,"b1");
	b2=create_sem(1,"b2");
	resume_thread(spawn_thread((thread_entry)show_screen,"screen updater",B_NORMAL_PRIORITY,NULL));

	if(gr_mode==MODE_MONO)
	{
		shiftmod = old_shiftmod = 2;
	}
	else
	{
		shiftmod = old_shiftmod = 0;
	}
	scr_width = scr_def[old_shiftmod].w;
	scr_height = scr_def[old_shiftmod].h;
	SM_B(MEM(0xff8260),2);
	OpenWindow();
}

static int beos_keytab[][2] =
{
	{ 0x02,ST_F1 },
	{ 0x03,ST_F2 },
	{ 0x04,ST_F3 },
	{ 0x05,ST_F4 },
	{ 0x06,ST_F5 },
	{ 0x07,ST_F6 },
	{ 0x08,ST_F7 },
	{ 0x09,ST_F8 },
	{ 0x0a,ST_F9 },
	{ 0x0b,ST_F10 },
	{ 0x01,ST_ESC },
	{ 0x12,ST_1 },
	{ 0x13,ST_2 },
	{ 0x14,ST_3 },
	{ 0x15,ST_4 },
	{ 0x16,ST_5 },
	{ 0x17,ST_6 },
	{ 0x18,ST_7 },
	{ 0x19,ST_8 },
	{ 0x1a,ST_9 },
	{ 0x1b,ST_0 },
	{ 0x1c,ST_MINUS },
	{ 0x1d,ST_EQUAL },
	{ 0x33,ST_BACKSL },
	{ 0x11,ST_GRAVE }, /* '_ */
	{ 0x46,ST_APOST }, 
	{ 0x1e,ST_BS },
	{ 0x26,ST_TAB },
	{ 0x27,ST_Q },
	{ 0x28,ST_W },
	{ 0x29,ST_E },
	{ 0x2a,ST_R },
	{ 0x2b,ST_T },
	{ 0x2c,ST_Y },
	{ 0x2d,ST_U },
	{ 0x2e,ST_I },
	{ 0x2f,ST_O },
	{ 0x30,ST_P },
	{ 0x31,ST_SQ_OPEN },	/* [ */
	{ 0x32,ST_SQ_CLOSE },	/* ] */
	{ 0x47,ST_RETURN },
	{ 0x5c,ST_CONTROL },
	{ 0x3c,ST_A },
	{ 0x3d,ST_S },
	{ 0x3e,ST_D },
	{ 0x3f,ST_F },
	{ 0x40,ST_G },
	{ 0x41,ST_H },
	{ 0x42,ST_J },
	{ 0x43,ST_K },
	{ 0x44,ST_L },
	{ 0x45,ST_SEMIC },
	{ 0x4b,ST_LSH },
	{ 0x4c,ST_Z },
	{ 0x4d,ST_X },
	{ 0x4e,ST_C },
	{ 0x4f,ST_V },
	{ 0x50,ST_B },
	{ 0x51,ST_N },
	{ 0x52,ST_M },
	{ 0x53,ST_COMMA },
	{ 0x54,ST_DOT },
	{ 0x55,ST_SLASH },
	{ 0x56,ST_RSH },
	{ 0x5d,ST_ALT },
	{ 0x5e,ST_SPACE },
	{ 0x3b,ST_CAPSLOCK },
	{ 0x64,ST_KP_0 },
	{ 0x58,ST_KP_1 },
	{ 0x59,ST_KP_2 },
	{ 0x5a,ST_KP_3 },
	{ 0x48,ST_KP_4 },
	{ 0x49,ST_KP_5 },
	{ 0x4a,ST_KP_6 },
	{ 0x37,ST_KP_7 },
	{ 0x38,ST_KP_8 },
	{ 0x39,ST_KP_9 },
	{ 0x57,ST_UP },
	{ 0x61,ST_LEFT },
	{ 0x63,ST_RIGHT }, 
	{ 0x62,ST_DOWN }, 
	{ 0x1f,ST_INSERT },
	{ 0x34,ST_DELETE },
	{ 0x5b,ST_KP_ENTER },
	{ 0x3a,ST_KP_PLUS },
	{ 0x25,ST_KP_MINUS },
	{ 0x24,ST_KP_MULT },
	{ 0x65,ST_KP_DOT },
	{ 0x23,ST_KP_DIV },
	{ 0x0e,ST_HELP },
	{ 0x10,ST_UNDO },
	{ 0x21,ST_KP_OPEN },
	{ 0x36,ST_KP_CLOSE },
	{ 0x5f,ST_CAPSLOCK },
	{ 0x60,ST_CONTROL },
	{ 0x66,ST_ISO }
};

char kstat[95];

extern int mb,mx,my;
extern unsigned char *kstates;

static void process_events(void)
{	
	int i;
	if (mouse_update())
	{	ikbd_pointer(mx,my,639,399);
		ikbd_button(1, (mb & 1) != 0);
		ikbd_button(2, (mb & 2) != 0);
	}
	keyboard_update();

	if((kstates[0x0c>>3] & (1<<(7-(0x0c%8))))) { mc68000_reset(); return; }

	for(i=0;i<95;i++)
	{	
		if((kstates[beos_keytab[i][0]>>3] & (1<<(7-(beos_keytab[i][0]%8)))))
		{
			if(kstat[i]==0)
			{
				ikbd_key(beos_keytab[i][1],1);
				kstat[i]=1;
			}
		}
		else
		{
			if(kstat[i]!=0)
			{
				ikbd_key(beos_keytab[i][1],0);
				kstat[i]=0;
			}
		}

	}
}

void x_screen_shifter (void)
{
	int i, w, pal = 0;

	process_events();

	acquire_sem(b2);
		
	if (old_shiftmod != shiftmod)
	{
		old_shiftmod = shiftmod;
		change_mode();
	}

	release_sem(b1);

	for (i=0; i<16; i++)
	{
		if ((w = LM_UW(&color[i])) != old_color[i])
		{
			pal = 1;	
			old_color[i] = w;
		}
	}
	
	if (pal) change_colors();
}

/* This is the keycode translation table. It is dependent on your kind of X
 * terminal, and on the mapping you want to use for the ST keys.
 * To make your own table, you'll need to examine the `xmodmap -pke' output
 * on your system, and write the corresponding ST keycode as the item with
 * the index of the keycode you want it on in the array below.
 *
 * Maybe, sometime in the future, a translation table based on Keysyms would
 * be a better idea...
 */
void x_init_keys(void)
{
	int i;
	for(i=0;i<95;i++)
	{
		kstat[i]=0;
	}		
}

